"""
File: genotype.py
Description: Instantiate a GT file object.
CreateDate: 2023/10/26
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Union, List
from typing_extensions import Literal
from io import TextIOWrapper
from os import getcwd
from os.path import abspath
from warnings import filterwarnings
from re import sub
from tqdm import tqdm
from natsort import natsort_key
from pandas import Series, DataFrame, read_table, read_excel, concat
from matplotlib.pyplot import rcParams, figure, tick_params, savefig
from seaborn import heatmap
from click import echo
from pybioinformatic.task_manager import TaskManager
from pybioinformatic.biopandas import read_file_as_dataframe_from_stdin, interval_stat
filterwarnings("ignore")


# =====================================================================================================================#
# NOTICE:                                                                                                              #
# The "__check_hom" and "__get_all_allele" functions cannot be defined in the "GenoType" class.                        #
# Because the "GenoType.parallel_stat_MHM" method runs with multiprocessing,                                           #
# the subprocess cannot call the private method of the "GenoType" class.                                               #
# =====================================================================================================================#
def __check_hom(row: Series) -> Series:
    """Check which samples are homozygous genotypes at specific loci."""
    genotype_set = {'AT', 'TA',
                    'AG', 'GA',
                    'AC', 'CA',
                    'GC', 'CG',
                    'GT', 'TG',
                    'CT', 'TC'}
    data = [str(value) not in genotype_set and '/' not in str(value) and str(value) != ''
            for value in row]
    return Series(data, row.index)


def __get_all_allele(row: Series) -> str:
    """Get all allele at specified loci."""
    all_allele = ''
    ref = row[0]
    for value in row[1:]:
        if '/' in str(value):
            left, right = str(value).split('/')[0], str(value).split('/')[1]
            if 'ins' in left:
                all_allele += 'I'
            elif 'del' in left:
                all_allele += 'D'
            elif left == '-':
                all_allele += ref[0]
            else:
                all_allele += left[0]
            if 'ins' in right:
                all_allele += 'I'
            elif 'del' in right:
                all_allele += 'D'
            elif right == '-':
                all_allele += ref[0]
            else:
                all_allele += right[0]
        elif 'ins' in str(value):
            all_allele += 'II'
        elif 'del' in str(value):
            all_allele += 'DD'
        elif str(value) == '-':
            all_allele += ref[0] * 2
        else:
            value = str(value).replace('*', 'U')
            all_allele += value
    return all_allele


def stat_MHM(df: DataFrame) -> DataFrame:
    """
    Calculate MissRate, HetRate, and MAF (MHM) from specified DataFrame.
    The top 4 columns of DataFrame must be SNP ID, chromosome, position, and reference sequence info respectively,
    and the header of DataFrame is not None.
    """
    snp_ref = df.iloc[:, :4]
    # Calculate MissRate
    sample_num = len(df.columns.tolist()) - 4
    df['MissRate(%)'] = df.isnull().sum(axis=1) / sample_num * 100
    # Calculate HetRate
    df.fillna('', inplace=True)  # fill NA
    df['HetRate(%)'] = df.iloc[:, 4:-1].apply(lambda row: (1 - __check_hom(row).sum() / (row != '').sum()) * 100, axis=1)
    # Calculate MAF
    df['all_gt'] = df.iloc[:, 3:-2].apply(lambda row: __get_all_allele(row), axis=1)
    df['total'] = df['all_gt'].apply(len)
    df['A'] = df['all_gt'].str.count('A') / df['total']
    df['G'] = df['all_gt'].str.count('G') / df['total']
    df['C'] = df['all_gt'].str.count('C') / df['total']
    df['T'] = df['all_gt'].str.count('T') / df['total']
    df['D'] = df['all_gt'].str.count('D') / df['total']
    df['I'] = df['all_gt'].str.count('I') / df['total']
    df['U'] = df['all_gt'].str.count('U') / df['total']
    df['MAF'] = df.loc[:, ['A', 'G', 'C', 'T', 'D', 'I', 'U']].apply(lambda row: sorted(row, reverse=True)[1], axis=1)
    # Output results
    df = df.loc[:, ['MissRate(%)', 'HetRate(%)', 'MAF']]
    merge = snp_ref.join(df)
    return merge


class GenoType:
    """
    Standard genotype file object.
    """

    def __init__(self, path: Union[str, TextIOWrapper]):
        self.name = abspath(path) if isinstance(path, str) else abspath(path.name.replace('<', '').replace('>', ''))
        self.__open = open(path) if isinstance(path, str) else path

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.__open.close()
        except AttributeError:
            pass

    @staticmethod
    def __allele_sort(allele: str) -> str:
        if '/' in allele:
            return '/'.join(sorted(allele.split('/')))
        elif 'ins' in allele or 'del' in allele:
            return allele
        else:
            return ''.join(sorted(allele))

    def allele_sort(self, df: DataFrame) -> DataFrame:
        snp_ref = df.iloc[:, :4]
        sorted_allele = df.iloc[:, 4:].applymap(self.__allele_sort, na_action='ignore')
        df = snp_ref.join(sorted_allele)
        return df

    def to_dataframe(self,
                     sheet: Union[str, int, List[Union[str, int]]] = 0,
                     index_col: int = None,
                     sort_allele: bool = True) -> DataFrame:
        try:  # read from text file
            if 'gz' in self.name:
                df = read_table(self.name, index_col=index_col)
            else:
                df = read_table(self.__open, index_col=index_col)
        except UnicodeDecodeError:  # read from Excel file
            df = read_excel(self.name, sheet, index_col=index_col)
        if index_col:
            df.sort_index(key=natsort_key, inplace=True)  # sort by index
        else:
            df.sort_values([df.columns[1], df.columns[2]], key=natsort_key, inplace=True)  # sort by values
        if sort_allele:
            df = self.allele_sort(df)
        return df

    def to_dataframes(self,
                      sheet: Union[str, int, List[Union[str, int]]] = 0,
                      index_col: int = None,
                      chunk_size: int = 10000) -> DataFrame:
        if 'stdin' in self.name:  # read from stdin
            dfs = read_file_as_dataframe_from_stdin(chunk_size=chunk_size, index_col=index_col)
        elif self.name.endswith('gz'):  # read from compressed text (xx.gz) file
            dfs = read_table(self.name, chunksize=chunk_size, index_col=index_col)
        else:
            try:  # read from text file
                dfs = read_table(self.__open, chunksize=chunk_size, index_col=index_col)
            except UnicodeDecodeError:  # read from Excel file
                echo('\033[33mWarning: Excel file cannot be processed in chunks.\033[0m')
                return self.to_dataframe(sheet, index_col)
        return dfs

    def parallel_stat_MHM(self, num_processing: int):
        """Calculate the MissRate, HetRate and MAF (MHM) of SNP sites from GT files parallely."""
        dfs = self.to_dataframes()
        if not isinstance(dfs, DataFrame):  # Calculate with multiprocessing
            params = ((df,) for df in dfs)  # Index of DataFrame is 0, 1, 2, ...
            tkm = TaskManager(num_processing=num_processing, params=params)
            ret = tkm.parallel_run_func(stat_MHM)
            stat_dfs = [i.get() for i in ret]
            # Merge results of each multiprocessing
            stat_df = concat(stat_dfs)
        else:
            stat_df = stat_MHM(dfs)
        stat_df.sort_values(stat_df.columns.tolist()[1:3], inplace=True, key=natsort_key)
        stat_df.fillna(0, inplace=True)
        return stat_df

    def merge(self, other,
              how: Literal['inner', 'outer'],
              sheet1: Union[str, int, List[Union[str, int]]] = None,
              sheet2: Union[str, int, List[Union[str, int]]] = None,
              ):
        df1 = self.to_dataframe(sheet=sheet1)
        df1[df1.columns[1]] = df1[df1.columns[1]].astype(str)
        df1[df1.columns[2]] = df1[df1.columns[2]].astype(str)
        df1['tmp'] = df1[df1.columns[1]] + '_' + df1[df1.columns[2]] + '_' + df1[df1.columns[3]]
        df1.set_index('tmp', drop=True, inplace=True)
        df1 = df1.iloc[:, 4:]

        df2 = other.to_dataframe(sheet=sheet2)
        df2[df2.columns[1]] = df2[df2.columns[1]].astype(str)
        df2[df2.columns[2]] = df2[df2.columns[2]].astype(str)
        df2['tmp'] = df2[df2.columns[1]] + '_' + df2[df2.columns[2]] + '_' + df2[df2.columns[3]]
        df2.set_index('tmp', drop=True, inplace=True)
        df2 = df2.iloc[:, 4:]

        merge = df1.join(other=df2, how=how)
        samples = merge.columns.tolist()
        samples.sort(key=natsort_key)
        merge = merge.loc[:, samples]
        merge.insert(0, 'Chr', [i.split('_')[0] for i in merge.index.tolist()])
        merge.insert(1, 'Pos', [i.split('_')[1] for i in merge.index.tolist()])
        merge.insert(2, 'Ref', [i.split('_')[2] for i in merge.index.tolist()])
        merge.index.name = 'ID'
        merge.rename(index=lambda i: '_'.join(i.split('_')[:2]), inplace=True)
        merge.fillna('NA', inplace=True)
        merge.sort_index(key=natsort_key, inplace=True)
        return merge

    def __draw_consistency_heatmap(
            self,
            consistency_df: DataFrame,
            cmap: str = "crest",
            output_path: str = getcwd(),
            font_name: str = 'Arial'
    ) -> None:
        # Set font and dpi.
        dpi = 300
        if len(consistency_df) <= 40:
            font_size = 8
        elif 40 < len(consistency_df) <= 100:
            font_size = 6
        elif 100 < len(consistency_df) <= 150:
            font_size = 4
        else:
            font_size = 1
            dpi = 1000
        rcParams['pdf.fonttype'] = 42
        rcParams['font.family'] = font_name
        rcParams['font.size'] = font_size
        # Set figure attributions.
        figure(figsize=(15, 10))
        # Plot heatmap.
        ax = heatmap(
            data=consistency_df,
            cmap=cmap,
            linecolor='w',
            linewidths=0.5,
            xticklabels=True,
            yticklabels=True,
            cbar_kws={'shrink': 0.4}
        )
        tick_params('both', length=0)  # Set scale length.
        # Set color bar ticks and ticks label.
        min_gs = consistency_df.min(numeric_only=True).min() + 1
        max_gs = consistency_df.max(numeric_only=True).max()
        min_value = int(min_gs)
        max_value = int(max_gs)
        step = int((max_value - min_value) / 10)
        step = 1 if step == 0 else step
        cbar = ax.collections[0].colorbar
        cbar_ticks = [i for i in range(min_value, max_value + step, step) if i <= 100]
        cbar.set_ticks(cbar_ticks)
        cbar.set_ticklabels(cbar_ticks, fontsize=8)
        cbar.ax.tick_params(width=0.3)
        # # Save figure.
        savefig(f'{output_path}/Consistency.heatmap.pdf', bbox_inches='tight', dpi=dpi)

    def self_compare(self, other,
                     sheet1: Union[str, int, List[Union[str, int]]] = None,
                     sheet2: Union[str, int, List[Union[str, int]]] = None,
                     output_path: str = getcwd()) -> None:
        """Genotype consistency of different test batches in a single sample."""
        # Read in gt.
        df1 = self.to_dataframe(sheet1, index_col=0, sort_allele=False)
        df1[df1.columns[0]] = df1[df1.columns[0]].astype(str)
        df1[df1.columns[1]] = df1[df1.columns[1]].astype(str)
        df1['tmp'] = df1[df1.columns[0]] + '_' + df1[df1.columns[1]] + '_' + df1[df1.columns[2]]
        df1.set_index('tmp', drop=True, inplace=True)
        df2 = other.to_dataframe(sheet2, index_col=0, sort_allele=False)
        df2[df2.columns[0]] = df2[df2.columns[0]].astype(str)
        df2[df2.columns[1]] = df2[df2.columns[1]].astype(str)
        df2['tmp'] = df2[df2.columns[0]] + '_' + df2[df2.columns[1]] + '_' + df2[df2.columns[2]]
        df2.set_index('tmp', drop=True, inplace=True)
        # Check whether the two GT files contain the same loci.
        if df1.index.tolist() != df2.index.tolist():
            echo('\033[31mError: The two GT file loci to be compared are inconsistent.\033[0m', err=True)
            exit()
        # Calculate genotype consistency of each sample under different test batches.
        samples = set(df1.columns[3:]) & set(df2.columns[3:])
        data = []
        with tqdm(total=len(samples), unit='sample') as pbar:
            for sample in samples:
                NA_site = set(df1[sample][df1[sample].isnull()].index) | set(df2[sample][df2[sample].isnull()].index)
                NA_num = len(NA_site)
                series1 = df1[sample][~df1[sample].isnull()]
                series2 = df2[sample][~df2[sample].isnull()]
                IdenticalCount = series1.eq(series2).sum()
                TotalCount = len(df1) - NA_num
                GS = IdenticalCount / TotalCount * 100
                data.append([sample, IdenticalCount, NA_num, TotalCount, GS])
                pbar.update(1)
        bins = range(0, 105, 5)  # Set the consistency statistics interval.
        sample_consistency = DataFrame(data, columns=['SampleName', 'IdenticalCount', 'NaCount', 'TotalCount', 'GS(%)'])
        sample_consistency.sort_values('SampleName', key=natsort_key, inplace=True)
        sample_consistency.fillna(0.000, inplace=True)
        interval_stat_df = interval_stat(ser=sample_consistency['GS(%)'], bins=bins, precision=0, name='Count')
        # Write results to output file.
        sample_consistency.to_csv(f'{output_path}/Sample.consistency.xls', sep='\t', index=False)
        interval_stat_df.to_csv(f'{output_path}/Interval.stat.xls', sep='\t', header=False)

    def compare(self, other,
                sheet1: Union[str, int, List[Union[str, int]]] = None,
                sheet2: Union[str, int, List[Union[str, int]]] = None,
                cmap: str = "crest",
                output_path: str = getcwd(),
                font_name: str = 'Arial') -> None:
        """
        Calculate genotype consistency.
        Output Sample.consistency.xls, Sample.Consistency.xls and Compare.GT.xls three files.
        """
        # Step1: Read GT file as DataFrame.
        df1 = self.to_dataframe(sheet1)  # index = 0, 1, 2, ...
        df2 = other.to_dataframe(sheet2)  # index = 0, 1, 2, ...
        # Step2: Select the site intersection of two GT files.
        left_on = df1.columns[0]
        right_on = df2.columns[0]
        merge = df1.merge(df2, left_on=left_on,  # index = 0, 1, 2, ...
                          right_on=right_on)  # Avoid inconsistency between the two GT file ID fields
        # Step3: Calculate genotype consistency.
        df1_sample_num = len(df1.columns.tolist()) - 4
        left_sample_range = list(range(4, 4 + df1_sample_num))
        if df1.columns[0] == df2.columns[0]:
            right_sample_range = list(range(len(df1.columns) + 3, len(merge.columns)))
        else:
            right_sample_range = list(range(len(df1.columns) + 4, len(merge.columns)))
        consistency_df = DataFrame()
        sample_pair = set()
        data = []
        with tqdm(total=len(left_sample_range), unit='sample') as pbar:  # Show process bar
            for index1 in left_sample_range:
                gt1 = merge.iloc[:, index1]
                gt1_NA = set(gt1[gt1.isnull()].index)
                gt1.name = sub(r'_x\b', '', gt1.name)
                pbar.set_description(f'Processing {gt1.name}')
                for index2 in right_sample_range:
                    gt2 = merge.iloc[:, index2]
                    gt2_NA = set(gt1[gt2.isnull()].index)
                    NA_site_index = gt1_NA | gt2_NA
                    NA_num = len(NA_site_index)
                    TotalCount = len(merge) - NA_num
                    gt2.name = sub(r'_y\b', '', gt2.name)
                    gt1 = gt1[~gt1.isnull()]
                    gt2 = gt2[~gt2.isnull()]
                    IdenticalCount = gt1.eq(gt2).sum()
                    GS = 0.00 if TotalCount == 0 else '%.2f' % (IdenticalCount / TotalCount * 100)
                    data.append([gt1.name, gt2.name, IdenticalCount, NA_num, TotalCount, GS])
                    if gt1.name != gt2.name:
                        if (f'{gt1.name}-{gt2.name}' not in sample_pair) and \
                                (f'{gt2.name}-{gt1.name}' not in sample_pair):
                            consistency_df.loc[gt2.name, gt1.name] = GS
                            sample_pair.add(f'{gt1.name}-{gt2.name}')
                            sample_pair.add(f'{gt2.name}-{gt1.name}')
                    else:
                        consistency_df.loc[gt1.name, gt2.name] = ''
                pbar.update(1)
        header = ['Sample1', 'Sample2', 'IdenticalCount', 'NaCount', 'TotalCount', 'GS(%)']
        fmt1 = DataFrame(data, columns=header)
        fmt1.sort_values(['Sample1', 'GS(%)', 'Sample2'], key=natsort_key, inplace=True, ascending=[True, False, True])
        fmt1.to_csv(f'{output_path}/Sample.consistency.fmt1.xls', sep='\t', index=False)
        # Step4: Draw consistency heatmap.
        consistency_df.to_csv(f'{output_path}/Sample.consistency.fmt2.xls', sep='\t', na_rep='')
        self.__draw_consistency_heatmap(
            consistency_df=read_table(f'{output_path}/Sample.consistency.fmt2.xls', index_col=0),
            cmap=cmap,
            output_path=output_path,
            font_name=font_name
        )
        # Step5: Output GT file of test sample.
        right_sample_range.insert(0, 0)  # Only output site ID
        right_sample_range.insert(1, len(df1.columns) + 2)
        test_sample_gt_df = merge.iloc[:, right_sample_range]
        test_sample_gt_df.rename(columns=lambda i: sub(r'_[xy]\b', '', str(i)), inplace=True)
        test_sample_gt_df.sort_values(merge.columns[0], key=natsort_key, inplace=True)
        test_sample_gt_df.to_csv(f'{output_path}/Compare.GT.xls', sep='\t', index=False, na_rep='NA')



"""
File: biomatplotlib.py
Description: Drawing aid library for matplotlib.
CreateDate: 2024/7/18
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import List
from numpy import random
import matplotlib.colors as mcolors


def generate_unique_colors(num_colors: int) -> List[str]:
    """Randomly generates a specified number of non-redundant hexadecimal color codes."""

    def __generate_unique_colors(num_colors):
        colors = set()
        while len(colors) < num_colors:
            r, g, b = random.rand(3)
            color = (r, g, b)
            colors.add(color)
        return colors

    unique_colors = __generate_unique_colors(num_colors)
    hex_colors = [mcolors.to_hex(color) for color in unique_colors]
    return hex_colors



"""
File: gff.py
Description: Instantiate a GFF file object.
CreateDate: 2021/11/27
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from io import TextIOWrapper, StringIO
from typing import Union, List, Dict, Tuple, Generator
from os.path import abspath
from gzip import GzipFile
from click import echo
from pandas import DataFrame, Series, read_table
from natsort import natsorted
from pybioinformatic.fasta import Fasta
from pybioinformatic.sequence import Nucleotide
from pybioinformatic.task_manager import TaskManager


class Gff:
    def __init__(self, path: Union[str, TextIOWrapper]):
        if isinstance(path, str):
            self.name = abspath(path)
            if path.endswith('gz'):
                self.__open = GzipFile(self.name, 'rb')
                self.line_num = sum(1 for line in self.__open if not str(line, 'utf8').startswith('#'))
                self.__open.seek(0)
                self.anno_line_num = sum(1 for line in self.__open if str(line, 'utf8').startswith('#'))
                self.__open.seek(0)
            else:
                self.__open = open(path, encoding='utf8')
                self.line_num = sum(1 for line in self.__open if not line.startswith('#'))
                self.__open.seek(0)
                self.anno_line_num = sum(1 for line in self.__open if line.startswith('#'))
                self.__open.seek(0)
        else:
            self.name = abspath(path.name)
            if path.name == '<stdin>':
                string_io = StringIO()
                for line in path:
                    string_io.write(line)
                string_io.seek(0)
                self.__open = string_io
                self.line_num = sum(1 for line in self.__open if not line.startswith('#'))
                self.__open.seek(0)
                self.anno_line_num = sum(1 for line in self.__open if line.startswith('#'))
                self.__open.seek(0)
            else:
                if self.name.endswith('gz'):
                    self.__open = GzipFile(self.name, 'rb')
                    self.line_num = sum(1 for line in self.__open if not str(line, 'utf8').startswith('#'))
                    self.__open.seek(0)
                    self.anno_line_num = sum(1 for line in self.__open if str(line, 'utf8').startswith('#'))
                    self.__open.seek(0)
                else:
                    self.__open = path
                    self.line_num = sum(1 for line in self.__open if not line.startswith('#'))
                    self.__open.seek(0)
                    self.anno_line_num = sum(1 for line in self.__open if line.startswith('#'))
                    self.__open.seek(0)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.__open.close()
        except AttributeError:
            pass

    def __seek_zero(self):
        try:
            self.__open.seek(0)
        except AttributeError:
            pass

# Basic method==========================================================================================================
    def parse(self) -> Generator[Tuple[str, str, str, str, str, str, str, str, Dict[str, str]], None, None]:
        """Parse information of each column of GFF file line by line."""
        for line in self.__open:
            line = str(line, 'utf8') if isinstance(line, bytes) else line
            if not line.startswith('#') and line.strip():
                split = line.strip().split('\t')
                chr_num, source, feature = split[0], split[1], split[2]
                start, end, score, strand, frame = split[3], split[4], split[5], split[6], split[7]
                attr_list = [attr for attr in split[8].split(';') if '=' in attr]
                attr_dict: Dict[str, str] = {attr.split('=')[0]: attr.split('=')[1] for attr in attr_list if attr}
                yield chr_num, source, feature, start, end, score, strand, frame, attr_dict
        self.__seek_zero()

    def to_dataframe(self) -> DataFrame:
        def parse_attribute(row: Series):
            attr_list = row['Attribute'].split(';')
            for attr in attr_list:
                k, v = attr.split('=')
                row[k] = v
            return row
        names = ['Chromosome', 'Source', 'Feature', 'Start', 'End', 'Score', 'Strand', 'Frame', 'Attribute']
        if self.name.endswith('gz'):
            df = read_table(self.name, header=None, names=names, dtype=str, comment='#')
        else:
            df = read_table(self.__open, header=None, names=names, dtype=str, comment='#')
        df[['Start', 'End']] = df[['Start', 'End']].astype(int)
        df = df.apply(func=parse_attribute, axis=1)
        df.drop(labels='Attribute', inplace=True, axis=1)
        raw = set(names[:-1])
        attr = list(set(df.columns) - raw)
        attr.sort()
        columns = names[:-1]
        columns.extend(attr)
        df = df.loc[:, columns]
        self.__seek_zero()
        return df  # index = [0, 1, 2, ...]

    def __check_feature(self, feature: str) -> Tuple[bool, str]:
        """Check whether specified feature is included in GFF file."""
        features = set(line[2] for line in self.parse())
        if feature in features:
            return True, f'"{feature}" have found.'
        else:
            return False, f'"{feature}" not found.'

    def to_dict(self,
                feature_type: str = None,
                anno: list = None
                ) -> Dict[str, List[Dict[str, Union[str, int]]]]:
        """Save the feature information in the GFF file into the dictionary."""
        # gff_dict = {
        #             Chr_num: [{ID: str, start: int, end: int, strand: str, ...}, {}, ...],
        #             Chr_num: [{}, {}, ...], ...
        #             }
        is_in_gff, msg = self.__check_feature(feature_type)
        if not is_in_gff:
            echo(f'\033[31mError: {msg}\033[0m', err=True)
            exit()
        gff_dict = {}
        for line in self.parse():
            if line[2] == feature_type or feature_type is None:
                if anno is not None:
                    item = {name: value for name, value in line[8].items() if name in anno}
                else:
                    item = {name: value for name, value in line[8].items()}
                pos = {'start': int(line[3]), 'end': int(line[4]), 'strand': line[6]}
                item.update(pos)
                if line[0] in gff_dict:
                    gff_dict[line[0]].append(item)
                else:
                    gff_dict[line[0]] = [item]
        return gff_dict

    def get_mRNA_dict(self) -> Dict[str, List[Dict[str, Union[str, int]]]]:
        """Get mRNA dict. The start and end based on mRNA length, not based on chromosome length."""
        mRNA_dict = {}  # {mRNA_id: [{feature_type: str, start: int, end: int, strand: str}, {}, ...], ...}
        mRNA_id = None
        mRNA_start = 0
        mRNA_len = 0
        for line in self.parse():
            if line[2] == 'mRNA':
                mRNA_id = line[8]['ID']
                mRNA_start = int(line[3])
                mRNA_len = int(line[4]) - int(line[3]) + 1
                if mRNA_id not in mRNA_dict:
                    mRNA_dict[mRNA_id] = []
                else:
                    echo(f'\033[31mError: The GFF file has repeat id {mRNA_id}.\033[0m')
                    exit()
            elif line[2] == 'CDS' or 'UTR' in line[2]:
                if line[8]['Parent'] == mRNA_id:
                    if line[6] == '-':
                        start = mRNA_len - (int(line[4]) - mRNA_start) - 1
                        end = mRNA_len - (int(line[3]) - mRNA_start) - 1
                        item = {'feature_type': line[2], 'start': start, 'end': end, 'strand': line[6]}
                    else:
                        item = {'feature_type': line[2], 'start': int(line[3]) - mRNA_start,
                                'end': int(line[4]) - mRNA_start, 'strand': line[6]}
                    mRNA_dict[mRNA_id].append(item)
                else:
                    echo('\033[31mError: GFF file is not sorted by mRNA ID.\033[0m', err=True)
                    exit()
        return mRNA_dict

    def summary(self) -> str:
        """A peek at the genome."""
        df = self.to_dataframe()
        df['Length'] = df['End'] - df['Start'] + 1
        content = [f'# Feature\tTotal\tMin_len\tMax_len\tMedian_len\tMean_len\n']
        features = df['Feature'].drop_duplicates().values
        for feature in features:
            total = len(df.loc[df['Feature'] == feature])
            min_len = df.loc[df['Feature'] == feature, 'Length'].min()
            max_len = df.loc[df['Feature'] == feature, 'Length'].max()
            median_len = '%.0f' % df.loc[df['Feature'] == feature, 'Length'].median()
            mean_len = '%.0f' % df.loc[df['Feature'] == feature, 'Length'].mean()
            content.append(f'{feature}\t{total}\t{min_len}\t{max_len}\t{median_len}\t{mean_len}\n')
        content = ''.join(content)
        return content

# GFF file sorted by id method==========================================================================================
    @staticmethod
    def __sort(line: str) -> tuple:
        split = line.strip().split('\t')
        chromosome = split[0]
        start, end = int(split[3]), int(split[4])
        attr_list = split[8].replace('=', ';').split(';')
        ID = attr_list[attr_list.index('ID') + 1]
        return chromosome, ID, start, end

    def sort(self) -> str:
        """Sort naturally in sequence according to chromosome, gene id, transcript id, start, and end."""
        if self.name.endswith('gz'):
            l = [
                str(line, 'utf8').strip()
                for line in self.__open
                if not str(line, 'utf8').startswith('#') and line.strip()
            ]
        else:
            l = [
                line.strip()
                for line in self.__open
                if not line.startswith('#') and line.strip()
            ]
        sorted_gff = natsorted(l, key=self.__sort)
        if not isinstance(self.__open, list):
            self.__open.seek(0)
        return '\n'.join(sorted_gff)

# Sequence extraction method============================================================================================
    @staticmethod
    def _extract_seq(nucl_obj: Nucleotide, features: list, feature_id_set: set, id_filed: str = 'ID'):
        sub_seq_list = []
        for feature in features:  # feature = {ID: str, start: int, end: int, strand: str, ...}
            sub_seq_obj = nucl_obj[feature['start'] - 1:feature['end']]
            anno = list(set(feature.keys()) - {'start', 'end', 'strand'})
            anno.sort()
            ID = [f'{i}={feature[i].replace(" ", "_")}' for i in anno if i != id_filed]
            ID.insert(0, feature[id_filed])
            ID = ' '.join(ID)
            sub_seq_obj.id = ID
            if feature_id_set and feature[id_filed] in feature_id_set:
                sub_seq_list.append(sub_seq_obj)
            elif not feature_id_set:
                sub_seq_list.append(sub_seq_obj)
        return sub_seq_list

    def extract_seq(self,
                    fasta_file: Union[str, TextIOWrapper],
                    feature_type: str = 'gene',
                    id_field: str = 'ID',
                    feature_id_set: set = None,
                    remain_attr: list = None,
                    num_processing: int = None) -> Nucleotide:
        """Extract sequences of specified feature type from GFF file."""
        is_in_gff, msg = self.__check_feature(feature_type)
        if not is_in_gff:
            echo(f'\033[31mError: {msg}\033[0m', err=True)
            exit()
        gff_dict = self.to_dict(feature_type, remain_attr)
        params = []
        with Fasta(fasta_file) as fa:
            for nucl_obj in fa.parse():
                try:
                    features = gff_dict[nucl_obj.id]  # features = [{feature1}, {feature2}, ...]
                except KeyError:
                    pass  # Some sequences (eg. scaffold, contig) may not have annotation
                else:
                    params.append((nucl_obj, features, feature_id_set, id_field))

        tkm = TaskManager(num_processing=num_processing, params=params)
        rets = tkm.parallel_run_func(func=self._extract_seq)
        feature_seq_list = []
        for ret in rets:
            feature_seq_list.extend(ret.get())
        feature_seq_list.sort(key=lambda item: item.id)
        yield from feature_seq_list

    def miRNA_extraction(self) -> Nucleotide:
        """Extract miRNA sequence from GFF file."""
        for line in self.parse():
            attr_dict = line[8]
            seq_id = attr_dict['ID']
            seq = attr_dict['seq']
            yield Nucleotide(seq_id, seq)

# File format conversion method=========================================================================================
    def to_gtf(self) -> Generator[str, None, None]:
        """Convert the file format from GFF to GTF."""
        last_line = None
        gene_id = transcript_id = None
        i = 0
        for line in self.parse():
            i += 1
            current_line = list(line)
            if current_line[2] == 'gene':
                if last_line:
                    yield '\t'.join(last_line[:8]) + f'''\tgene_id "{gene_id}"; transcript_id "{transcript_id}";'''
                    last_line = None
                    gene_id = current_line[8]['ID']
                else:
                    gene_id = current_line[8]['ID']
                yield '\t'.join(current_line[:8]) + f'''\tgene_id "{gene_id}";'''
            elif current_line[2] == 'mRNA' or current_line[2] == 'transcript':
                current_line[2] = 'transcript'
                if last_line:
                    if gene_id is not None:
                        yield '\t'.join(last_line[:8]) + f'''\tgene_id "{gene_id}"; transcript_id "{transcript_id}";'''
                    else:
                        yield '\t'.join(last_line[:8]) + f'''\ttranscript_id "{transcript_id}";'''
                    last_line = None
                    transcript_id = current_line[8]['ID']
                else:
                    transcript_id = current_line[8]['ID']
                if gene_id is not None:
                    yield '\t'.join(current_line[:8]) + f'''\tgene_id "{gene_id}"; transcript_id "{transcript_id}";'''
                else:
                    yield '\t'.join(current_line[:8]) + f'''\ttranscript_id "{transcript_id}";'''
            elif 'UTR' in current_line[2] or 'CDS' in current_line[2]:
                current_line[2] = 'exon'
                current_line[7] = '.'
                if last_line:
                    if last_line[8]['Parent'] == current_line[8]['Parent'] == transcript_id:
                        if int(last_line[4]) + 1 == int(current_line[3]):
                            current_line[3] = last_line[3]
                            last_line = current_line
                        else:
                            if gene_id is not None:
                                yield ('\t'.join(last_line[:8]) +
                                       f'''\tgene_id "{gene_id}"; transcript_id "{transcript_id}";''')
                            else:
                                yield '\t'.join(last_line[:8]) + f'''\ttranscript_id "{transcript_id}";'''
                            last_line = current_line
                else:
                    last_line = current_line
            if self.line_num == i:
                if gene_id is not None:
                    yield '\t'.join(last_line[:8]) + f'''\tgene_id "{gene_id}"; transcript_id "{transcript_id}";'''
                else:
                    yield '\t'.join(last_line[:8]) + f'''\ttranscript_id "{transcript_id}";'''

    def to_bed(self, feature_type: Union[str, list] = None) -> str:
        """Convert the file format from GFF to BED."""
        for line in self.parse():
            if feature_type:
                if line[2] == feature_type or line[2] in feature_type:
                    yield f"{line[0]}\t{int(line[3]) - 1}\t{line[4]}\t{line[8]['ID']}\t{line[7]}\t{line[6]}"
            elif not feature_type:
                yield f"{line[0]}\t{int(line[3]) - 1}\t{line[4]}\t{line[8]['ID']}\t{line[7]}\t{line[6]}"

    def to_gsds(self) -> str:
        """Convert the file format from GFF to GSDS."""
        transcript_id, transcript_start = None, 0
        for line in self.parse():
            line = list(line)
            if line[2] == 'mRNA':
                transcript_id = line[8]['ID']
                transcript_start = int(line[3])
            elif line[2] == 'CDS' or 'UTR' in line[2]:
                if line[8]['Parent'] == transcript_id:
                    line[3] = str(int(line[3]) - transcript_start)
                    line[4] = (int(line[4]) - transcript_start)
                    yield f"{transcript_id}\t{line[3]}\t{line[4]}\t{line[2]}\t{line[7]}"
                else:
                    echo(f'\033[33mWarning: The order of GFF file is wrong, '
                               f'this will cause some information to be lost.\033[0m', err=True)

# Feature density count=================================================================================================
    def get_feature_density(self,
                            chr_len_dict: Dict[str, int],
                            feature_type: str = 'gene',
                            span: int = 100000) -> Generator[str, None, None]:
        """Get feature density."""
        is_in_gff, msg = self.__check_feature(feature_type)
        if not is_in_gff:
            echo(f'\033[31mError: {msg}\033[0m', err=True)
            exit()
        if min(list(chr_len_dict.values())) / span < 1:
            echo('\033[33mError: Density statistical interval is too large.\033[0m', err=True)
            exit()
        gff_dataframe = self.to_dataframe()
        gff_dataframe['Site'] = (gff_dataframe['Start'] + gff_dataframe['End']) / 2
        for chr_num, length in chr_len_dict.items():
            df = gff_dataframe[(gff_dataframe['Chromosome'] == chr_num) & (gff_dataframe['Feature'] == feature_type)]
            sites = df['Site']
            for i in range(span, length, span):
                count = len(sites[(sites <= i) & (sites >= i - span + 1)])
                yield f'{chr_num}\t{i - span + 1}\t{i}\t{count}'
            if length % span != 0:
                count = len(sites[(sites <= length) & (sites >= length // span * span + 1)])
                yield f'{chr_num}\t{length // span * span + 1}\t{length}\t{count}'



"""
File: util.py
Description: Util module.
CreateDate: 2024/5/12
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Callable
from natsort import natsort_key


class FuncDict(dict):
    def sort_by_keys(self, key: Callable = natsort_key,  reverse: bool = False):
        sorted_dict = FuncDict({key: self[key] for key in sorted(self, key=key, reverse=reverse)})
        self.clear()
        self.update(sorted_dict)

    def __getitem__(self, key):
        """Returns the same value as the key when a non-existent key is accessed."""
        try:
            return super().__getitem__(key)
        except KeyError:
            return key

    def __add__(self, other):
        for k, v in other.items():
            if k in self:
                self[k] += v
            else:
                self[k] = v



#!/usr/bin/env python
"""
File: base.py
Description: Next-generation sequencing (NGS) data analysis module.
CreateDate: 2025/4/18
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Dict
from os import makedirs
from os.path import abspath
from shutil import which
from click import echo
from pybioinformatic import FuncDict


def parse_sample_info(sample_info: str) -> Dict[str, list]:
    d = {}  # {sample_name: [fastq_file1, fastq_file2, comparative_combination], ...}
    with open(sample_info) as f:
        for line in f:
            split = line.strip().split('\t')
            d[split[0]] = split[1:4]
    return d


def build_ref_index(
        fasta_file: str,
        bwa_exe: str = None,
        gatk_exe: str = None,
        samtools_exe: str = None,
        hisat2_build_exe: str = None,
        bowtie_build: str = None,
        bowtie2_build: str = None
) -> str:
    software_list = [
        which(i)
        for i in [bwa_exe, gatk_exe, samtools_exe, hisat2_build_exe, bowtie_build, bowtie2_build]
        if i is not None and which(i)
    ]
    if not software_list:
        echo('\033[33mWaring: No any valid executable file was specified, ignore building any index of fasta file.\033[0m', err=True)
        exit()
    fasta_file = abspath(fasta_file)
    cmd_dict = {
        'bwa': f' index {fasta_file} {fasta_file}',
        'gatk': f' CreateSequenceDictionary -R {fasta_file}',
        'samtools': f' faidx {fasta_file}',
        'hisat2-build': f' {fasta_file} {fasta_file}',
        'bowtie-build': f' {fasta_file} {fasta_file}',
        'bowtie2-build': f' {fasta_file} {fasta_file}',
    }
    try:
        cmd = '\n'.join(
            [
                i + cmd_dict[i.split('/')[-1]]
                for i in software_list
            ]
        )
    except KeyError as s:
        echo(f'\033[31mError: Invalid command {s}.\033[0m', err=True)
        exit()
    return cmd


class NGSAnalyser:
    """
    \033[32mNext-generation sequencing (NGS) data analysis pipeline.\033[0m

    \033[34m:param read1: Pair end read1 fastq file.\033[0m
    \033[34m:param read2: Pair end read2 fastq file.\033[0m
    \033[34m:param ref_genome: Reference genome fasta file.\033[0m
    \033[34m:param output_path: Output path.\033[0m
    \033[34m:param num_threads: Number of threads.\033[0m
    \033[34m:param sample_name: Sample name.\033[0m
    """

    _exe_set = {'fastp', 'bwa', 'samtools', 'hisat2', 'gatk', 'bowtie', 'bowtie2'}

    def __init__(
            self,
            read1: str,
            read2: str,
            ref_genome: str,
            output_path: str,
            num_threads: int = 10,
            sample_name: str = None,
            exe_path_dict: Dict[str, str] = None
    ):
        if exe_path_dict is None:
            exe_path_dict = {}
        ## Input file
        self.read1 = abspath(read1)
        self.read2 = abspath(read2)
        self.ref_genome = abspath(ref_genome)
        self.output_path = abspath(output_path)
        self.num_threads = num_threads
        self.sample_name = self.read1.split('/')[-1].split('.')[0] if sample_name is None else sample_name
        self.exe_path_dict = FuncDict(
            {
                k: which(v)
                for k, v in exe_path_dict.items()
                if k in self._exe_set and which(v) is not None
            }
        )
        ## Key output file
        # QC results file
        self.qc_path = f'{self.output_path}/01.QC/{self.sample_name}'
        self.read1_clean = f'{self.qc_path}/{self.sample_name}_1_clean.fq.gz'
        self.read2_clean = f'{self.qc_path}/{self.sample_name}_2_clean.fq.gz'
        self.fastp_json = f'{self.qc_path}/{self.sample_name}.fastp.json'
        self.fastp_html = f'{self.qc_path}/{self.sample_name}.fastp.html'
        # bwa mem results file
        self.mapping_path = f'{self.output_path}/02.mapping/{self.sample_name}'
        self.bwa_mem_raw_bam = f'{self.mapping_path}/{self.sample_name}.bwa.mem.sort.bam'
        self.bwa_mem_filtered_bam = f'{self.mapping_path}/{self.sample_name}.bwa.mem.sort.map30.bam'
        self.bwa_mem_markdup_bam = f'{self.mapping_path}/{self.sample_name}.bwa.mem.sort.map30.markdup.bam'
        # hisat2 results file
        self.hisat2_raw_bam = f'{self.mapping_path}/{self.sample_name}.ht2.sort.bam'

    @staticmethod
    def other_options(cmd: str, other_options: dict):
        other_options = ' '.join([f'{k} {v}' for k, v in other_options.items()])
        cmd += f' {other_options}'
        return cmd

    def run_fastp(
            self,
            read1_clean: str = None,
            read2_clean: str = None,
            **other_options
    ):
        fastp = which(self.exe_path_dict['fastp'])
        read1_clean = abspath(read1_clean) if read1_clean else self.read1_clean
        read2_clean = abspath(read2_clean) if read2_clean else self.read2_clean
        fastp_json = '/'.join(read1_clean.split('/')[:-1]) + f'/{self.sample_name}.fastp.json' if read1_clean else self.fastp_json
        fastp_html = '/'.join(read1_clean.split('/')[:-1]) + f'/{self.sample_name}.fastp.html' if read1_clean else self.fastp_html
        makedirs(self.qc_path, exist_ok=True)
        cmd = (
            f'{fastp} -w {self.num_threads} '
            f'-i {self.read1} -I {self.read2} '
            f'-o {read1_clean} '
            f'-O {read2_clean} '
            f'-j {fastp_json} '
            f'-h {fastp_html}'
        )
        return self.other_options(cmd, other_options) if other_options else cmd

    def run_hisat2(
            self,
            read1_clean: str = None,
            read2_clean: str = None,
            out_bam: str = None,
            **other_options
    ):
        hisat2 = which(self.exe_path_dict['hisat2'])
        samtools = which(self.exe_path_dict['samtools'])
        read1_clean = abspath(read1_clean) if read1_clean else self.read1_clean
        read2_clean = abspath(read2_clean) if read2_clean else self.read2_clean
        out_bam = abspath(out_bam) if out_bam else self.hisat2_raw_bam
        makedirs(self.mapping_path, exist_ok=True)
        hisat2_cmd = (
            f'{hisat2} -p {self.num_threads} '
            f'-x {self.ref_genome} '
            f'-1 {read1_clean} '
            f'-2 {read2_clean} '
            f'--summary-file  {self.mapping_path}/{self.sample_name}.ht2.log'
        )
        hisat2_cmd = self.other_options(hisat2_cmd, other_options) if other_options else hisat2_cmd
        cmd = f'{hisat2_cmd} | {samtools} sort -@ {self.num_threads} -T {self.sample_name} - -o {out_bam}'
        return cmd

    def run_bwa_mem(
            self,
            read1_clean: str = None,
            read2_clean: str = None,
            out_bam: str = None,
            **other_options
    ) -> str:
        bwa = which(self.exe_path_dict['bwa'])
        samtools = which(self.exe_path_dict['samtools'])
        read1_clean = abspath(read1_clean) if read1_clean else self.read1_clean
        read2_clean = abspath(read2_clean) if read2_clean else self.read2_clean
        out_bam = abspath(out_bam) if out_bam else self.bwa_mem_raw_bam
        makedirs(self.mapping_path, exist_ok=True)
        bwa_cmd = (
            fr'{bwa} mem -t {self.num_threads} '
            fr'-R "@RG\tID:{self.sample_name}\tSM:{self.sample_name}\tPL:ILLUMINA" '
            fr'{self.ref_genome} {read1_clean} {read2_clean}'
        )
        bwa_cmd = self.other_options(bwa_cmd, other_options) if other_options else bwa_cmd
        cmd = f'{bwa_cmd} | {samtools} sort -@ {self.num_threads} -T {self.sample_name} -o {out_bam}'
        return cmd

    def filter_reads_by_mapQ(self, bam_file: str = None, map_q: int = 30) -> str:
        if map_q not in range(1, 60):
            echo(f'\033[31mError: Invalid mapping Q value: "{map_q}". It must be an integer between 0 and 60.\033[0m', err=True)
            exit()
        samtools = which(self.exe_path_dict['samtools'])
        makedirs(self.mapping_path, exist_ok=True)
        if bam_file:
            bam_file = abspath(bam_file)
            out_bam = f"{self.mapping_path}/{self.sample_name}{bam_file.split('/')[-1].replace(self.sample_name, '').replace('bam', 'map30.bam')}"
        else:
            bam_file = self.bwa_mem_raw_bam
            out_bam = self.bwa_mem_filtered_bam
        awk = r'''awk '{if($1~/@/){print}else{if( $7 == "=" &&  $5 >= %s ){print $0}}}' ''' % map_q
        cmd = f'{samtools} view -h {bam_file} | {awk}| samtools view -bS -T {self.ref_genome} - -o {out_bam}'
        return cmd

    def mark_duplicates(
            self,
            bam_file: str = None,
            **other_options
    ) -> str:
        gatk = which(self.exe_path_dict['gatk'])
        samtools = which(self.exe_path_dict['samtools'])
        makedirs(self.mapping_path, exist_ok=True)
        if bam_file:
            bam_file = abspath(bam_file)
            out_bam = f"{self.mapping_path}/{self.sample_name}{bam_file.split('/')[-1].replace(self.sample_name, '').replace('bam', 'markdup.bam')}"
            out_metrics = f"{self.mapping_path}/{self.sample_name}{bam_file.split('/')[-1].replace(self.sample_name, '')}.metrics"
        else:
            bam_file = self.bwa_mem_filtered_bam
            out_bam = self.bwa_mem_markdup_bam
            out_metrics = f'{self.bwa_mem_markdup_bam}.metrics'
        cmd = (
            f'{gatk} MarkDuplicates '
            f'-I {bam_file} '
            f'-M {out_metrics} '
            f'-O {out_bam}'
        )
        cmd = self.other_options(cmd, other_options) if other_options else cmd
        cmd += f'\n{samtools} index {out_bam}'
        return cmd

    def stats_depth(self, bam_file: str = None, **other_options) -> str:
        samtools = which(self.exe_path_dict['samtools'])
        makedirs(self.mapping_path, exist_ok=True)
        if bam_file:
            bam_file = abspath(bam_file)
            out_file = f"{self.mapping_path}/{self.sample_name}{bam_file.split('/')[-1].replace(self.sample_name, '')}.depth"
        else:
            bam_file = self.bwa_mem_markdup_bam
            out_file = f"{self.bwa_mem_markdup_bam.split('/')[-1]}.depth"
        cmd = f'{samtools} depth {bam_file} -o {self.mapping_path}/{out_file}'
        return self.other_options(cmd, other_options) if other_options else cmd



from pybioinformatic.NGS.base import parse_sample_info, build_ref_index, NGSAnalyser
from pybioinformatic.NGS.DNA_seq import GatkSNPCalling
from pybioinformatic.NGS.RNA_seq import (
    RNASeqAnalyser,
    LncRNAPredictor,
    LncRNATargetPredictor,
    LncRNAClassification
)


__all__ = [
    'parse_sample_info',
    'build_ref_index',
    'NGSAnalyser',
    'GatkSNPCalling',
    'RNASeqAnalyser',
    'LncRNAPredictor',
    'LncRNATargetPredictor',
    'LncRNAClassification'
]



"""
File: sequence.py
Description: Instantiate a sequence object (including nucleotide and amino acid sequences).
CreateDate: 2021/11/26
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Tuple, Union
from collections import Counter
from re import findall, sub
from natsort import natsort_key
from ViennaRNA import fold, circfold, RNA
from click import echo


class Sequence:
    def __init__(self, seq_id: str, sequence: str, strip: bool = True):
        self.id = sub(r'^[>@]', '', seq_id)
        if strip:
            self.seq = sequence.replace('\n', '')
        else:
            self.seq = sequence
        if '*' in sequence:
            self.len = len(sequence.replace('\n', '').replace('*', ''))
        else:
            self.len = len(sequence.replace('\n', ''))

    def __str__(self) -> str:
        if 'length' not in self.id:
            return f'>{self.id} length={len(self)}\n{self.seq.strip()}'
        else:
            self.id = sub(r'length=\d+', f'length={len(self)}', self.id)
            return f'>{self.id}\n{self.seq.strip()}'

    def __contains__(self, item) -> bool:
        """
        Define when implement "Sequence_obj1 in Sequence_obj2",
        if Sequence_obj1.seq in Sequence_obj2.seq, return True,
        otherwise return False.
        """
        return True if item.seq.replace('\n', '') in self.seq.replace('\n', '') else False

    def __ne__(self, other) -> bool:
        """
        Define when implement "Sequence_obj1 != Sequence_obj2",
        if Sequence_obj1.seq != Sequence_obj2.seq, return True,
        otherwise return False.
        """
        return True if self.seq.replace('\n', '') != other.seq.replace('\n', '') else False

    def __eq__(self, other) -> bool:
        """
        Define when implement "Sequence_obj1 == Sequence_obj2",
        if Sequence_obj1.seq == Sequence_obj2.seq, return True,
        otherwise return False.
        """
        return True if self.seq.replace('\n', '') == other.seq.replace('\n', '') else False

    def __hash__(self):
        return hash(self.seq)

    def __lt__(self, other) -> bool:
        """
        Define when implement "Sequence_obj1 < Sequence_obj2",
        if len(Sequence_obj1.seq) < len(Sequence_obj2.seq),
        return True, otherwise return False.
        """
        return True if len(self) < len(other) else False

    def __le__(self, other) -> bool:
        """
        Define when implement "Sequence_obj <= Sequence_obj2",
        if len(Sequence_obj1.seq) <= len(Sequence_obj2.seq),
        return True, otherwise return False.
        """
        return True if len(self) <= len(other) else False

    def __gt__(self, other) -> bool:
        """
        Define when implement "Sequence_obj > Sequence_obj2",
        if len(Sequence_obj1.seq) > len(Sequence_obj2.seq),
        return True, otherwise return False.
        """
        return True if len(self) > len(other) else False

    def __ge__(self, other) -> bool:
        """
        Define when implement "Sequence_obj >= Sequence_obj2",
        if len(Sequence_obj1.seq) >= len(Sequence_obj2.seq),
        return True, otherwise return False.
        """
        return True if len(self) >= len(other) else False

    def __iter__(self) -> str:
        """ Implement "iter(self.seq)". """
        yield from self.seq

    def __getitem__(self, item):
        """
        Define when implement "Sequence_obj[int:int:int].seq",
        it is equal to "Sequence_obj.seq[int:int:int]",
        but the return value type is same as raw object.
        """
        if item.start is None or item.start == 0:
            start = 1
        else:
            start = item.start + 1
        if item.stop is None or item.stop > self.len:
            stop = self.len
        else:
            stop = item.stop
        if item.step is None:
            step = 1
        else:
            step = item.step
        if isinstance(self, Reads):
            return Reads(seq_id=f"{self.id} slice({start}:{stop}:{step})",
                         sequence=self.seq.replace('\n', '')[item],
                         desc=self.desc,
                         quality=self.quality[item])
        elif isinstance(self, Nucleotide):
            return Nucleotide(f"{self.id} slice({start}:{stop}:{step})", self.seq.replace('\n', '')[item])
        elif isinstance(self, Protein):
            return Protein(f"{self.id} slice({start}:{stop}:{step})", self.seq.replace('\n', '')[item])
        else:
            return Sequence(f"{self.id} slice({start}:{stop}:{step})", self.seq.replace('\n', '')[item])

    def k_mer(self, k: int):
        """Get K-mer sequence."""
        for i in range(0, len(self) - k + 1):
            yield self[i:i+k]

    def count_k_mer(self, k: int):
        """Count k-mer frequency."""
        k_mer_list = [obj.seq for obj in self.k_mer(k)]
        k_mer_count = Counter(k_mer_list)
        return k_mer_count

    def __len__(self) -> int:
        return len(self.seq.replace('*', '').replace('\n', ''))

    def get_seq_len_info(self) -> str:
        """Get sequence length information."""
        return f'{self.id}\t{self.len}'

    def find_motif(self, motif: str, only_forward: bool = False) -> str:
        """Find the motif in the sequence."""
        ret = []
        ret_str = f'{self.id} not found motif.'
        raw_seq = self.seq.replace('\n', '')
        matched = findall(rf'{motif}', raw_seq)
        matched = list(set(matched))
        if matched:
            for match in matched:
                if raw_seq == match:
                    ret.append(f"{self.id}\t1\t{len(raw_seq)}\t+\t{motif}\t{match}")
                else:
                    split = raw_seq.split(match)
                    end = 0
                    for i in split:
                        if i != split[-1]:
                            start = end + len(i) + 1
                            end = start + len(match) - 1
                            ret.append(f"{self.id}\t{start}\t{end}\t+\t{motif}\t{match}")
        if only_forward:
            ret.sort(key=natsort_key)
            return '\n'.join(ret) if ret else ret_str
        else:
            # If sequence is nucleotide, scan motif of reverse complementary strand.
            # If sequence is peptide chain, scan motif of reverse strand.
            try:
                seq = ''
                # Sequence is peptide chain
                if set(self.seq.upper()) - {'A', 'G', 'C', 'T', 'U', 'N'}:
                    seq = raw_seq
                # Sequence is nucleotide
                elif 'T' in self.seq or 't' in self.seq:
                    for n in raw_seq:
                        seq += Nucleotide.dna_complementary_dict[n]
                else:
                    for n in raw_seq:
                        seq += Nucleotide.rna_complementary_dict[n]
                seq = seq[::-1]
                matched = findall(rf'{motif}', seq)
                matched = list(set(matched))
                if matched:
                    for match in matched:
                        if seq == match:
                            ret.append(f"{self.id}\t1\t{len(seq)}\t-\t{motif}\t{match}")
                        else:
                            split = seq.split(match)
                            end = 0
                            for i in split:
                                if i != split[-1]:
                                    start = end + len(i) + 1
                                    end = start + len(match) - 1
                                    e = len(seq) - start + 1
                                    s = len(seq) - end + 1
                                    ret.append(f"{self.id}\t{s}\t{e}\t-\t{motif}\t{match}")
                    ret.sort(key=natsort_key)
            except AttributeError:
                pass
            return '\n'.join(ret) if ret else ret_str

    def display_set(self, n: int = 60):
        raw_seq = self.seq.replace('\n', '')
        seq: list = findall(r'\w{1,%s}' % n, raw_seq)
        if '*' in raw_seq:
            if len(seq[-1]) == n:
                seq.append('*')
            else:
                seq[-1] = seq[-1] + '*'
        if isinstance(self, Nucleotide):
            return Nucleotide(self.id, '\n'.join(seq) + '\n', False)
        elif isinstance(self, Nucleotide):
            return Protein(self.id, '\n'.join(seq) + '\n', False)


class Nucleotide(Sequence):
    dna_complementary_dict = {'A': 'T', 'T': 'A', 'G': 'C', 'C': 'G', 'N': 'N',
                              'a': 't', 't': 'a', 'g': 'c', 'c': 'g', 'n': 'n'}

    rna_complementary_dict = {'A': 'U', 'U': 'A', 'G': 'C', 'C': 'G', 'N': 'N',
                              'a': 'u', 'u': 'a', 'g': 'c', 'c': 'g', 'n': 'n'}

    codon_table = {'UUU': 'F', 'UUC': 'F', 'UUA': 'L', 'UUG': 'L',
                   'UCU': 'S', 'UCC': 'S', 'UCA': 'S', 'UCG': 'S',
                   'UAU': 'Y', 'UAC': 'Y', 'UAA': '*', 'UAG': '*',
                   'UGU': 'C', 'UGC': 'C', 'UGA': '*', 'UGG': 'W',
                   'CUU': 'L', 'CUC': 'L', 'CUA': 'L', 'CUG': 'L',
                   'CCU': 'P', 'CCC': 'P', 'CCA': 'P', 'CCG': 'P',
                   'CAU': 'H', 'CAC': 'H', 'CAA': 'Q', 'CAG': 'Q',
                   'CGU': 'R', 'CGC': 'R', 'CGA': 'R', 'CGG': 'R',
                   'AUU': 'I', 'AUC': 'I', 'AUA': 'I', 'AUG': 'M',
                   'ACU': 'T', 'ACC': 'T', 'ACA': 'T', 'ACG': 'T',
                   'AAU': 'N', 'AAC': 'N', 'AAA': 'K', 'AAG': 'K',
                   'AGU': 'S', 'AGC': 'S', 'AGA': 'R', 'AGG': 'R',
                   'GUU': 'V', 'GUC': 'V', 'GUA': 'V', 'GUG': 'V',
                   'GCU': 'A', 'GCC': 'A', 'GCA': 'A', 'GCG': 'A',
                   'GAU': 'D', 'GAC': 'D', 'GAA': 'E', 'GAG': 'E',
                   'GGU': 'G', 'GGC': 'G', 'GGA': 'G', 'GGG': 'G'}

    def __add__(self, other):
        """
        Define when implement "Nucleotide_obj1 + Nucleotide_obj2",
        it is equal to "Nucleotide_obj1.seq + Nucleotide_obj2.seq",
        but no value is returned, just the seq and len attributes of Nucleotide_obj1 are changed.
        """
        self.seq = self.seq + other.seq
        self.len = len(self)

    def __neg__(self):
        """
        Define when implement "-Nucleotide_obj", it is equal to "Nucleotide_obj.get_reverse_complementary_seq()".
        """
        return self.get_reverse_complementary_seq()

    @staticmethod
    def random_nucl(name: str = None, length: Union[int, list, tuple] = None, bias: Union[float, list, tuple] = 1.0):
        """
        Generate a random nucleotide sequence.
        :param name: Name of random nucleotide sequence. {type: str}
        :param length: Length of random nucleotide sequence. {type: int, list, or tuple; default: (100, 1000)}
        :param bias: Base preference. {type: float, 4-list, or 4-tuple; default: 1}
        :return: Nucleotide class instance.
        """
        import random
        if name is None:
            seed = list('abcdefghijklmnopqrstuvwxyz0123456789'.upper())
            name = ''.join([random.choice(seed) for _ in range(10)])
        if length is None:
            length = random.randint(100, 1000)
        elif isinstance(length, (list, tuple)):
            length = random.randint(length[0], length[1])
        if isinstance(bias, float):
            bias = [bias for _ in range(4)]
        random_nucl_seq = random.choices(population=list('AGCT'), weights=bias, k=length)
        return Nucleotide(name, ''.join(random_nucl_seq))

    def get_reverse_complementary_seq(self):
        """Get reverse complementary sequence of DNA or RNA."""
        seq = ''
        if 'T' in self.seq or 't' in self.seq:
            for n in self.seq.replace('\n', ''):
                seq += self.dna_complementary_dict[n]
        else:
            for n in self.seq.replace('\n', ''):
                seq += self.rna_complementary_dict[n]
        return Nucleotide(f"{self.id} reverse_complementary_chain", seq[::-1])

    def base_count(self) -> Union[Tuple[float, float, float, float, float, str], None]:
        """Get the percentage content of four bases."""
        A = self.seq.upper().count('A') / len(self) * 100
        G = self.seq.upper().count('G') / len(self) * 100
        C = self.seq.upper().count('C') / len(self) * 100
        T = self.seq.upper().count('T') / len(self) * 100
        U = self.seq.upper().count('U') / len(self) * 100
        N = self.seq.upper().count('N') / len(self) * 100
        a = '%.2f%s' % (A, '%')
        g = '%.2f%s' % (G, '%')
        c = '%.2f%s' % (C, '%')
        t = '%.2f%s' % (T, '%')
        u = '%.2f%s' % (U, '%')
        n = '%.2f%s' % (N, '%')
        if 'U' in self.seq.upper() and 'T' not in self.seq.upper():
            summary = f"Base content statistics of {self.id}\nA: {a}\nG: {g}\nC: {c}\nU: {u}\nN: {n}"
            return A, G, C, U, N, summary
        elif 'T' in self.seq.upper() and 'U' not in self.seq.upper():
            summary = f"Base content statistics of {self.id}\nA: {a}\nG: {g}\nC: {c}\nT: {t}\nN: {n}"
            return A, G, C, T, N, summary
        else:
            echo('\033[31mError: Unknown nucleotide sequence, both "U" and "T" in sequence.\033[0m', err=True)

    def find_SSR(self) -> str:
        """Find the simple sequence repeat (SSR) in the DNA sequence."""
        ret = []
        raw_seq = self.seq.replace('\n', '').replace('U', 'T')
        pattern = r'((\w+)\2\2+)'
        matched = findall(pattern, raw_seq)
        matched = list(set(matched))
        if matched:
            for match_tuple in matched:
                for match in match_tuple:
                    if len(match) >= 6:
                        try:
                            ssr_unit = max(findall(r'(\w+)\1{2}', match))
                            if len(set([nucl for nucl in ssr_unit])) == 1:
                                ssr_unit = list(set([nucl for nucl in ssr_unit]))[0]
                        except ValueError:
                            pass
                        else:
                            count = match.count(ssr_unit)
                            split = raw_seq.split(match)
                            end = 0
                            for i in split:
                                if i != split[-1]:
                                    start = end + len(i) + 1
                                    end = start + len(match) - 1
                                    ret.append(f"{self.id}\t{start}\t{end}\t({ssr_unit}){count}\t{match}")
                                ret.sort(key=lambda item: (int(item.split('\t')[1]), int(item.split('\t')[2])))
            return '\n'.join(ret) if ret else f'{self.id} not found SSR.'
        else:
            return f'{self.id} not found SSR.'

    def translation(self, complete: bool = True):
        """Translate nucleotide sequence to peptide chain."""
        peptide_chain = []
        append = peptide_chain.append
        for i in range(0, len(self), 3):
            codon = self.seq.upper().replace('T', 'U').replace('\n', '')[i:i + 3]
            try:
                append(self.codon_table[codon])
            except KeyError:
                append('-')
        peptide_chain = ''.join(peptide_chain)
        if complete:
            try:
                peptide_chain = (str(max(findall(r'M[A-Z]+\*', peptide_chain), key=len)))
                return Protein(f"{self.id} peptide_chain", peptide_chain)
            except ValueError:
                return Protein(f"{self.id} peptide_chain", '')
        else:
            peptide_chain = str(max(findall(r'M?[A-Z]+\*?', peptide_chain), key=len))
            return Protein(f"{self.id} peptide_chain", peptide_chain)

    def ORF_prediction(self, min_len: int = 1, complete: bool = True, only_plus: bool = False):
        """
        ORF prediction.
        :param min_len: minimal ORF length (type=int) {default=1}
        :param complete: whether consider ORF integrity (type=bool) {default=True}
        :param only_plus: whether only consider plus chain (type=bool) {default=False}
        :return: the longest ORF (type=Protein)
        """
        plus1 = self.translation(complete)
        plus2 = self[1:].translation(complete)
        plus3 = self[2:].translation(complete)
        prot_obj_list = [plus1, plus2, plus3]
        if set(self.seq.upper()) - {'A', 'G', 'C', 'T'}:
            msg = f'\033[33mWaring: "{self.id}" has degenerate bases, the peptide chain translation will stop at the degenerate base.\033[0m'
            echo(msg, err=True)
        if not only_plus:
            minus1 = (-self).translation(complete)
            minus2 = (-self)[1:].translation(complete)
            minus3 = (-self)[2:].translation(complete)
            prot_obj_list.extend([minus1, minus2, minus3])
        try:
            longest_ORF = max(prot_obj_list, key=len)
            if len(longest_ORF) >= min_len:
                longest_ORF.id = f'{self.id} ORF_prediction'
                return longest_ORF
            else:
                return f"{self.id} not found ORF."
        except ValueError:
            return f"{self.id} not found ORF."

    def circular_translation(self) -> tuple:
        """Translate a nucleotide sequence circularly."""
        for i in range(len(self)):
            cds = pep = ''
            j = i
            while True:
                if j + 3 <= len(self):
                    codon = self.seq[j:j + 3].upper()
                    if cds:
                        cds += codon
                        pep += self.codon_table[codon.replace('T', 'U')]
                        if self.codon_table[codon.replace('T', 'U')] == '*':
                            j += 3
                            break
                    else:
                        if codon == 'ATG':
                            cds += codon
                            pep += self.codon_table[codon.replace('T', 'U')]
                        else:
                            break
                else:
                    if j < len(self):
                        n = len(self) - j
                        codon = self.seq[-n:].upper() + self.seq[:3 - n].upper()
                        if cds:
                            cds += codon
                            pep += self.codon_table[codon.replace('T', 'U')]
                        else:
                            if codon == 'ATG':
                                cds += codon
                                pep += self.codon_table[codon.replace('T', 'U')]
                            else:
                                break
                        if self.codon_table[codon.replace('T', 'U')] == '*':
                            j += 3
                            break
                    else:
                        if (j - i) == len(self):
                            break
                        else:
                            if len(self) - j % len(self) < 3:
                                n = len(self) - j % len(self)
                                codon = self.seq[-n:].upper() + self.seq[:3 - n].upper()
                            else:
                                codon = self.seq[j % len(self):j % len(self) + 3].upper()
                            cds += codon
                            pep += self.codon_table[codon.replace('T', 'U')]
                            if self.codon_table[codon.replace('T', 'U')] == '*':
                                j += 3
                                break
                j += 3
            if cds and pep.endswith('*'):
                if j > len(self):
                    cds = Nucleotide(
                        f'{self.id} circular_translation_cds start={i + 1} end={j % len(self)} circular={j // len(self)}',
                        cds)
                    pep = Protein(
                        f'{self.id} circular_translation_pep start={i + 1} end={j % len(self)} circular={j // len(self)}',
                        pep)
                else:
                    cds = Nucleotide(f'{self.id} circular_translation_cds start={i + 1} end={j} circular=0', cds)
                    pep = Protein(
                        f'{self.id} circular_translation_pep start={i + 1} end={j} circular=0', pep)
                yield cds, pep

    def predict_secondary_structure(self, ps_file: str, circular: bool = False):
        seq = self.seq.replace('\n', '')
        ss, mfe = circfold(seq) if circular else fold(seq)
        RNA.PS_rna_plot(seq, ss, ps_file)
        if self.seq.count('\n') > 1:
            l1 = self.seq.strip().split('\n')
            n = len(l1[0])
            l2 = findall(r'[(.)]{1,%s}' % n, ss)
            s = f'>{self.id}\n'
            for i in zip(l1, l2):
                s += f'{i[0]}\n{i[1]}\n'
            return s.strip(), mfe
        else:
            return f'{self}\n{ss}', mfe


class Protein(Sequence):
    @staticmethod
    def random_prot(name: str = None, length: Union[int, list, tuple] = None, complete: bool = True):
        """
        Generate a random protein sequence.
        :param name: Name of random protein sequence. {type: str}
        :param length: Length of random protein sequence. {type: int, list or tuple; default: (100, 150)}
        :param complete: Whether random protein sequence is completed. {type: bool; default: True}
        :return: Protein class
        """
        import random
        if name is None:
            seed = list('abcdefghijklmnopqrstuvwxyz0123456789'.upper())
            name = ''.join([random.choice(seed) for _ in range(10)])
        if length is None:
            length = random.randint(100, 150)
        elif isinstance(length, (list, tuple)):
            length = random.randint(length[0], length[1])
        if complete:
            random_prot_seq = 'M' + ''.join([random.choice(list('HYRKCVQIGFTSWPMEADLN')) for _ in range(length - 1)]) + '*'
        else:
            random_prot_seq = ''.join([random.choice(list('HYRKCVQIGFTSWPMEADLN')) for _ in range(length)])
        return Protein(name, random_prot_seq)


class Reads(Nucleotide):
    def __init__(self, seq_id: str, sequence: str, desc: str, quality: str):
        super(Reads, self).__init__(seq_id, sequence)
        self.desc = desc
        self.quality = quality
        quality_dict = {}
        quality_str = '!"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~'
        for i in range(len(quality_str)):
            quality_dict[quality_str[i]] = i
        self.__quality_dict = quality_dict

    def __str__(self) -> str:
        if 'length' not in self.id:
            return f'@{self.id} length={len(self)}\n{self.seq.strip()}\n{self.desc}\n{self.quality}'
        else:
            self.id = sub(r'length=\d+', f'length={len(self)}', self.id)
            return f'@{self.id}\n{self.seq.strip()}\n{self.desc}\n{self.quality}'

    def get_quality(self) -> Tuple[int, ...]:
        t = tuple(self.__quality_dict[quality] for quality in self.quality)
        return t

    def filter_N(self, rate: float = 0.1, n: int = None) -> bool:
        if rate > 1:
            echo(f'\033[31mError: Invalid rate value "{rate}", it must be between 0 and 1.', err=True)
            exit()
        seq = self.seq.upper()
        if n is None:
            if seq.count('N') / self.len >= rate:
                return False
            else:
                return True
        else:
            if (seq.count('N') / self.len >= rate) or (seq.count('N') >= n):
                return False
            else:
                return True


# Test
if __name__ == '__main__':
    DNA = Nucleotide.random_nucl(length=1000)
    DNA = DNA.display_set()
    print(DNA)
    print(DNA.find_motif('(AAGCT){1,}'))
    print(DNA.base_count()[-1])
    print(DNA.ORF_prediction(complete=False).display_set())

    rev_com_DNA = DNA.get_reverse_complementary_seq().display_set(80)
    print(rev_com_DNA)
    print(rev_com_DNA.ORF_prediction().display_set())



#!/usr/bin/env python
"""
File: task_manager.py
Description: Instance a TaskManager.
CreateDate: 2023/9/8
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Union, Iterable, Callable, Generator, Any
from io import TextIOWrapper
from subprocess import run, PIPE
from datetime import datetime
from getpass import getuser
from socket import gethostname
from multiprocessing import Pool
from click import echo


class TaskManager:
    def __init__(self,
                 commands: Iterable[str] = None,
                 num_processing: int = None,
                 params: Iterable[tuple] = None,
                 log_file: Union[str, TextIOWrapper] = None):
        try:
            self.task = list(commands)
        except TypeError:
            self.task = []
        self.params = params
        self.num_processing = num_processing
        if isinstance(log_file, str):
            self.loger = open(log_file, 'w')
        else:
            self.loger = log_file

    def add_task(self, command: Union[str, Iterable[str]]):
        if isinstance(command, str):
            self.task.append(command)
        else:
            command = list(command)
            self.task.extend(command)

    def del_task(self, index: Union[int, str] = None):
        if isinstance(index, int):
            del self.task[index]
        elif isinstance(index, str):
            self.task.remove(index)
        else:
            self.task.pop()

    def clear_task(self):
        self.task = []

    def echo_and_exec_cmd(self, cmd: str, show_cmd: bool = True) -> Union[None, str]:
        """
        NOTICE: This method cannot run normally when it is called by multiprocessing.
                Because different subprocess will share same TaskManager.loger attribution.
        """
        if show_cmd:
            echo(
                f'\033[33m[{getuser()}@{gethostname()}: '
                f'{datetime.now().replace(microsecond=0)}]\n$ '
                f'\033[0m\033[36m{cmd}\033[0m', self.loger, err=True
            )
        stdout = run(cmd, shell=True, executable="/bin/bash", stdout=PIPE)
        stdout = stdout.stdout.decode('utf-8')
        return stdout

    def serial_run_cmd(self) -> Generator[str, Any, None]:
        for cmd in self.task:
            echo(f'\033[33m[{getuser()}@{gethostname()}: {datetime.now().replace(microsecond=0)}]\n'
                 f'$ \033[0m\033[36m{cmd}\033[0m', self.loger, err=True)
            stdout = run(cmd, shell=True, executable="/bin/bash", stdout=PIPE)
            stdout = stdout.stdout.decode('utf-8')
            yield stdout

    def parallel_run_cmd(self):
        if not self.task:
            echo('\033[31mError: TaskManager has no task.\033[0m', err=True)
            exit()
        results = []
        with Pool(self.num_processing) as pool:
            for cmd in self.task:
                ret = pool.apply_async(self.echo_and_exec_cmd, args=(cmd,))
                results.append(ret)
            pool.close()
            pool.join()
        return results

    def parallel_run_func(self, func: Callable, call_back_func: Callable = None):
        results = []
        with Pool(self.num_processing) as pool:
            for param in self.params:
                ret = pool.apply_async(func, args=param, callback=call_back_func)
                results.append(ret)
            pool.close()
            pool.join()
        return results


if __name__ == '__main__':
    tkm = TaskManager()
    cmds = [f'echo Hello World x {i}' for i in range(10)]
    tkm.add_task(cmds)
    tkm.serial_run_cmd()



"""
File: DNA_seq.py
Description: DNA-seq analysis pipeline module.
CreateDate: 2025/4/18
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Dict
from os import makedirs
from os.path import abspath
from shutil import which
from pybioinformatic.NGS.base import NGSAnalyser


class GatkSNPCalling(NGSAnalyser):
    def __init__(
            self,
            read1: str,
            read2: str,
            ref_genome: str,
            output_path: str,
            num_threads: int = 10,
            sample_name: str = None,
            exe_path_dict: Dict[str, str] = None
    ):
        super().__init__(read1, read2, ref_genome, output_path, num_threads, sample_name, exe_path_dict)
        self.variant_path = f'{self.output_path}/03.variant/{self.sample_name}'
        self.gvcf = f"{self.variant_path}/{self.bwa_mem_markdup_bam.split('/')[-1]}.gvcf"
        self.vcf = f"{self.variant_path}/{self.bwa_mem_markdup_bam.split('/')[-1]}.vcf"
        self.filtered_vcf = f"{self.variant_path}/{self.bwa_mem_markdup_bam.split('/')[-1]}.filtered.vcf"

    def run_HaplotypeCaller(
            self,
            bam_file: str = None,
            **other_options
    ) -> str:
        gatk = which(self.exe_path_dict['gatk'])
        makedirs(self.variant_path, exist_ok=True)
        if bam_file:
            bam_file = abspath(bam_file)
            out_gvcf = f"{self.variant_path}/{self.sample_name}{bam_file.split('/')[-1].replace(self.sample_name, '')}.gvcf"
        else:
            bam_file = self.bwa_mem_markdup_bam
            out_gvcf = self.gvcf
        cmd = (f'{gatk} HaplotypeCaller '
               f'-ERC GVCF '
               f'-I {bam_file} '
               f'-R {self.ref_genome} '
               f'-O {out_gvcf}')
        return self.other_options(cmd, other_options) if other_options else cmd

    def run_GenotypeGVCFs(
            self,
            gvcf_file: str = None,
            **other_options
    ) -> str:
        gatk = which(self.exe_path_dict['gatk'])
        makedirs(self.variant_path, exist_ok=True)
        if gvcf_file:
            gvcf_file = abspath(gvcf_file)
            out_vcf = f"{self.variant_path}/{self.sample_name}{gvcf_file.split('/')[-1].replace(self.sample_name, '').replace('gvcf', 'vcf')}"
        else:
            gvcf_file = self.gvcf
            out_vcf = self.vcf
        cmd = (
            f'{gatk} GenotypeGVCFs '
            f'-R {self.ref_genome} '
            f'-V {gvcf_file} '
            f'-O {out_vcf}'
        )
        return self.other_options(cmd, other_options) if other_options else cmd

    def run_VariantFiltration(
            self,
            filter_expression: str = 'QD < 2.0 || MQ < 40.0 || FS > 60.0 || SOR > 3.0',
            vcf_file: str = None,
            **other_options
    ) -> str:
        gatk = which(self.exe_path_dict['gatk'])
        makedirs(self.variant_path, exist_ok=True)
        if vcf_file:
            vcf_file = abspath(vcf_file)
            out_vcf = f"{self.variant_path}/{self.sample_name}{vcf_file.split('/')[-1].replace(self.sample_name, '').replace('vcf', 'filtered.vcf')}"
        else:
            vcf_file = self.vcf
            out_vcf = self.filtered_vcf
        cmd = (
            f'{gatk} VariantFiltration '
            f'--filter-name  "HARD_TO_VALIDATE" '
            f'--filter-expression "{filter_expression}" '
            f'-R {self.ref_genome} '
            f'-V {vcf_file} '
            f'-O {out_vcf}'
        )
        return self.other_options(cmd, other_options) if other_options else cmd

    def pipeline(self) -> str:
        cmds = (
            f'{self.run_fastp()}\n\n'
            f'{self.run_bwa_mem()}\n\n'
            f'{self.filter_reads_by_mapQ()}\n\n'
            f'{self.mark_duplicates()}\n\n'
            f'{self.stats_depth()}\n\n'
            f'{self.run_HaplotypeCaller()}\n\n'
            f'{self.run_GenotypeGVCFs()}\n\n'
            f'{self.run_VariantFiltration()}'
        )
        return cmds



"""
File: RNA_seq.py
Description: RNA-seq analysis pipeline module.
CreateDate: 2024/7/24
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from io import TextIOWrapper
from typing import Literal, Union, Dict
from os import makedirs
from os.path import abspath, exists
from shutil import which
from tqdm import tqdm
from scipy.stats import pearsonr
from statsmodels.stats.multitest import multipletests
from pandas import DataFrame, Series, concat
from pybioinformatic.gtf import Gtf
from pybioinformatic.task_manager import TaskManager
from pybioinformatic.NGS import NGSAnalyser
from pybioinformatic.util import FuncDict


class RNASeqAnalyser(NGSAnalyser):
    def __init__(
            self,
            read1: str,
            read2: str,
            ref_genome: str,
            output_path: str,
            ref_gff: str = None,
            num_threads: int = 10,
            sample_name: str = None,
            exe_path_dict: Dict[str, str] = None
    ):
        self._exe_set.update({'stringtie'})
        super().__init__(read1, read2, ref_genome, output_path, num_threads, sample_name, exe_path_dict)
        self.ref_gff = abspath(ref_gff) if ref_gff else ref_gff
        self.assembly_path = f'{self.output_path}/03.assembly/{self.sample_name}'
        self.stringtie_gtf = f'{self.assembly_path}/{self.sample_name}.st.gtf'

    def run_stringtie(
            self,
            sorted_bam: str = None,
            out_gtf: str = None,
            **other_options
    ):
        stringtie = which(self.exe_path_dict['stringtie'])
        sorted_bam = abspath(sorted_bam) if sorted_bam else self.hisat2_raw_bam
        out_gtf = abspath(out_gtf) if out_gtf else self.stringtie_gtf
        makedirs(self.assembly_path, exist_ok=True)
        cmd = (
            f'{stringtie} '
            f'-p {self.num_threads} '
            f'-l {self.sample_name} '
            f'-o {out_gtf} '
            f'{sorted_bam}'
        )
        return self.other_options(cmd, other_options) if other_options else cmd

    def ssRNA_seq_pipeline(self):
        hisat2_options = {'--rna-strandness': 'RF'}
        stringtie_options = {'--rf': '', '-G': self.ref_gff} if self.ref_gff else {'--rf': ''}
        cmds = (
            f'{self.run_fastp()}\n\n'
            f'{self.run_hisat2(**hisat2_options)}\n\n'
            f'{self.run_stringtie(**stringtie_options)}'
        )
        return cmds


class LncRNAPredictor:
    _exe_set = {'CNCI.py', 'CPC2.py', 'PLEK', 'pfam_scan.pl', 'seqkit'}

    def __init__(
            self,
            nucl_fasta_file: str,
            output_path: str,
            num_thread: int = 10,
            module: Literal['ve', 'pl'] = 'pl',
            pep_fasta_file: str = None,
            PfamDatabase: str = None,
            exe_path_dict: Dict[str, str] = None
    ):
        if exe_path_dict is None:
            exe_path_dict = {}
        self.nucl = abspath(nucl_fasta_file)
        self.pep = abspath(pep_fasta_file) if pep_fasta_file else None
        self.PfamDatabase = abspath(PfamDatabase) if PfamDatabase else None
        self.output_path = abspath(output_path)
        self.num_thread = num_thread
        self.module = module
        self.exe_path_dict = FuncDict(
            {
                k: which(v)
                for k, v in exe_path_dict.items()
                if k in self._exe_set and which(v) is not None
            }
        )
        # output
        self.CNCI_out = f'{self.output_path}/CNCI/CNCI.index'
        self.CPC2_out = f'{self.output_path}/CPC2/CPC2.txt'
        self.PLEK_out = f'{self.output_path}/PLEK/PLEK.xls'
        self.PfamScan_out = f'{self.output_path}/PfamScan/pfamscan_out.txt'

    def run_CNCI(self):
        CNCI = which(self.exe_path_dict['CNCI.py'])
        cmd = f'{CNCI} -p {self.num_thread} -m {self.module} -f {self.nucl} -o {self.output_path}/CNCI'
        return cmd

    def run_CPC2(self):
        CPC2 = which(self.exe_path_dict['CPC2.py'])
        makedirs(f'{self.output_path}/CPC2', exist_ok=True)
        cmd = f'{CPC2} -i {self.nucl} --ORF TRUE -o {self.output_path}/CPC2/CPC2'
        return cmd

    def run_PLEK(self):
        PLEK = which(self.exe_path_dict['PLEK'])
        makedirs(f'{self.output_path}/PLEK', exist_ok=True)
        cmd = f'{PLEK} -thread {self.num_thread} -fasta {self.nucl} -out {self.PLEK_out}'
        return cmd

    def run_PfamScan(self):
        PfamScan = which(self.exe_path_dict['pfam_scan.pl'])
        makedirs(f'{self.output_path}/PfamScan', exist_ok=True)
        cmd = (f'{PfamScan} '
               f'-fasta {self.pep} '
               f'-dir {self.PfamDatabase} '
               f'-outfile {self.PfamScan_out}')
        return cmd

    def merge_results(self,
                      CNCI_results: str = None,
                      CPC2_results: str = None,
                      PLEK_results: str = None,
                      PfamScan_results: str = None):
        CNCI_results = abspath(CNCI_results) if CNCI_results else self.CNCI_out
        CPC2_results = abspath(CPC2_results) if CPC2_results else self.CPC2_out
        PLEK_results = abspath(PLEK_results) if PLEK_results else self.PLEK_out
        PfamScan_results = abspath(PfamScan_results) if PfamScan_results else self.PfamScan_out

        seqkit = which(self.exe_path_dict['seqkit'])
        CNCI_results = r'''awk -F'\t' '{if($2 == "noncoding"){print $1}}' %s | awk -F' ' '{print $1}' ''' % CNCI_results
        CPC2_results = r'''awk -F'\t' '{if($9 == "noncoding"){print $1}}' %s''' % CPC2_results
        PLEK_results = r'''awk -F'\t' '{if($1 == "Non-coding"){print $3}}' %s | sed 's/>//;s/ gene.*//' ''' % PLEK_results
        awk1 = r'''awk -F' ' '{if($1 == 3){print $2}}' '''
        awk2 = r'''awk -F' ' '{print $1}' %s''' % PfamScan_results
        seqkit_cmd1 = f'{seqkit} grep -f - {self.nucl}'
        seqkit_cmd2 = f'{seqkit} seq -m 200 -'
        cmd = (f'cat <({CNCI_results}) <({CPC2_results}) <({PLEK_results}) | sort | uniq -c | {awk1} | '
               f'grep -vFwf <({awk2}) - | {seqkit_cmd1} | {seqkit_cmd2} > {self.output_path}/lncRNA.fa')
        return cmd


class LncRNATargetPredictor:
    def __init__(self,
                 lncRNA_gtf_file: Union[str, TextIOWrapper] = None,
                 mRNA_gtf_file: Union[str, TextIOWrapper] = None,
                 distance: int = 100000,
                 lncRNA_exp_file: Union[str, TextIOWrapper] = None,
                 mRNA_exp_file: Union[str, TextIOWrapper] = None,
                 lncRNA_min_exp: float = 0.5,
                 mRNA_min_exp: float = 0.5,
                 r: float = 0.85,
                 FDR: float = 0.05,
                 q_value: float = 0.05,
                 num_processing: int = 1,
                 output_path: str = '.'):
        self.lncRNA_gtf = abspath(lncRNA_gtf_file)
        self.lncRNA_exp = abspath(lncRNA_exp_file)
        self.lncRNA_min_exp = lncRNA_min_exp
        self.mRNA_gtf = abspath(mRNA_gtf_file)
        self.mRNA_exp = abspath(mRNA_exp_file)
        self.mRNA_min_exp = mRNA_min_exp
        self.distance = distance
        self.r, self.FDR, self.q_value = r, FDR, q_value
        self.num_processing = num_processing
        self.output_path = abspath(output_path)

    @staticmethod
    def intersect(row: Series, mRNA: DataFrame, distance: int):
        lncRNA_chr = row['Chromosome']
        lncRNA_start, lncRNA_end = row['Start'], row['End']
        start = lncRNA_start - distance if lncRNA_start - distance > 0 else 1
        end = lncRNA_end + distance
        lncRNA_strand = row['Strand']
        candidate_mRNA = mRNA[
            (mRNA['Chromosome'] == lncRNA_chr) &
            (start <= mRNA['Start']) &
            (end >= mRNA['End'])
            ]
        candidate_mRNA['LncRNA_id'] = row.name
        candidate_mRNA['LncRNA_start'] = lncRNA_start
        candidate_mRNA['LncRNA_end'] = lncRNA_end
        candidate_mRNA['LncRNA_strand'] = lncRNA_strand
        return candidate_mRNA

    def co_location(self):
        with Gtf(self.lncRNA_gtf) as lncRNA_gtf, Gtf(self.mRNA_gtf) as mRNA_gtf:
            lncRNA_df = lncRNA_gtf.merge_by_transcript()
            mRNA_df = mRNA_gtf.merge_by_transcript()
            lncRNA_df.to_csv(f'{self.output_path}/lncRNA.bed', sep='\t')
            mRNA_df.to_csv(f'{self.output_path}/mRNA.bed', sep='\t')
            params = ((row[1], mRNA_df, self.distance) for row in lncRNA_df.iterrows())
            tkm = TaskManager(num_processing=self.num_processing, params=params)
            with tqdm(total=len(lncRNA_df), desc='Co-location predicting') as pbar:
                rets = tkm.parallel_run_func(func=self.intersect, call_back_func=lambda _: pbar.update())
            dfs = [ret.get() for ret in rets]
            lncRNA_mRNA_df = concat(dfs)
            lncRNA_mRNA_df.to_csv(f'{self.output_path}/co_loc.xls', sep='\t')

    @staticmethod
    def sub_process(x_name, x_data, y_name, y_data):
        r, p = pearsonr(x_data, y_data)
        return [x_name, y_name, r, p]

    def co_expression(self):
        lncRNA_exp_dict = {
            line.strip().split('\t')[0]: [float(i) for i in line.strip().split('\t')[1:]]
            for line in open(self.lncRNA_exp)
            if not line.startswith('Geneid') and line.strip() and
               all([float(i) >= self.lncRNA_min_exp for i in line.strip().split('\t')[1:]])
        }

        mRNA_exp_dict = {
            line.strip().split('\t')[0]: [float(i) for i in line.strip().split('\t')[1:]]
            for line in open(self.mRNA_exp)
            if not line.startswith('Geneid') and line.strip() and
               all([float(i) >= self.mRNA_min_exp for i in line.strip().split('\t')[1:]])
        }

        with tqdm(total=len(lncRNA_exp_dict) * len(mRNA_exp_dict), desc='Co-expression predicting') as pbar:
            params = ((k1, v1, k2, v2) for k1, v1 in lncRNA_exp_dict.items() for k2, v2 in mRNA_exp_dict.items())
            tkm = TaskManager(num_processing=self.num_processing, params=params)
            rets = tkm.parallel_run_func(func=self.sub_process, call_back_func=lambda _: pbar.update())

        data = [ret.get() for ret in rets]
        raw = DataFrame(data=data, columns=['lncRNA_id', 'mRNA_id', 'pcc', 'p_value'])
        raw.sort_values(by=['lncRNA_id', 'p_value'], inplace=True, ascending=[True, True])
        raw.to_csv(f'{self.output_path}/raw_co_exp.xls', sep='\t', index=False, float_format='%.12f', na_rep='NA')

        filter_df = raw[(raw['pcc'].abs() >= self.r) & (raw['p_value'] <= self.q_value)]
        filter_df['q_value'] = multipletests(pvals=filter_df['p_value'], alpha=self.FDR, method='fdr_bh')[1]
        filter_df = filter_df[filter_df['q_value'] <= self.q_value]
        filter_df.sort_values(by=['lncRNA_id', 'q_value'], inplace=True, ascending=[True, True])
        filter_df.to_csv(f'{self.output_path}/filter_co_exp.xls', sep='\t', index=False, float_format='%.12f', na_rep='NA')


class LncRNAClassification:
    def __init__(self, mRNA_gff_file: str, lncRNA_gtf_file: str, out_dir: str):
        self.mRNA_gff_file = abspath(mRNA_gff_file)
        self.mRNA_gtf_file = '.'.join(self.mRNA_gff_file.split('.')[:-1]) + '.gtf'
        self.lncRNA_gtf_file = abspath(lncRNA_gtf_file)
        out_dir = abspath(out_dir)
        makedirs(out_dir, exist_ok=True)
        self.gene_bed = f'{out_dir}/gene.bed'
        self.exon_bed = f'{out_dir}/exon.bed'
        self.intron_bed = f'{out_dir}/intron.bed'
        self.lncRNA_bed = f'{out_dir}/lncRNA.bed'
        self.sense = f'{out_dir}/sense.bed'
        self.antisense = f'{out_dir}/antisense.bed'
        self.intronic = f'{out_dir}/intronic.bed'
        self.intergenic = f'{out_dir}/intergenic.bed'

    def classification(self):
        bedtools = which('bedtools')
        gffread = which('gffread')
        gff2gtf = f'{gffread} {self.mRNA_gff_file} -T -o {self.mRNA_gtf_file}' if not exists(self.mRNA_gtf_file) else ''

        gene_bed = (r'''awk -F'\t' '{if($3 == "gene"){print $1,$4,$5,$9,$6,$7}}' OFS='\t' %s | sort -uV > %s''' %
                    (self.mRNA_gff_file, self.gene_bed))
        exon_bed = (r'''awk -F'\t' '{if($3 == "exon"){print $1,$4,$5,$9,$6,$7}}' OFS='\t' %s | sort -uV > %s''' %
                    (self.mRNA_gtf_file, self.exon_bed))
        intron_bed = f'{bedtools} subtract -a {self.gene_bed} -b {self.exon_bed} -s | sort -uV > {self.intron_bed}'
        lncRNA_bed = (r'''awk -F'\t' '{print $1,$4,$5,$9,$6,$7}' OFS='\t' %s | sort -uV > %s''' %
                      (self.lncRNA_gtf_file, self.lncRNA_bed))

        intronic = f'{bedtools} intersect -a {self.lncRNA_bed} -b {self.intron_bed} -s > {self.intronic}'

        sense = (
                r'''%s intersect -a <(awk '{if(match($0, /transcript_id "[a-zA-Z0-9]*.*[a-zA-Z0-9]*"/)) print substr($0, RSTART, RLENGTH)}' %s | sort -uV | grep -Fwvf - %s) -b %s -s > %s''' %
                (bedtools, self.intronic, self.lncRNA_bed, self.gene_bed, self.sense)
        )

        antisense = (
                r'''%s intersect -a <(cat %s %s | awk '{if(match($0, /transcript_id "[a-zA-Z0-9]*.*[a-zA-Z0-9]*"/)) print substr($0, RSTART, RLENGTH)}' | sort -uV | grep -Fwvf - %s) -b %s -S > %s''' %
                (bedtools, self.intronic, self.sense, self.lncRNA_bed, self.gene_bed, self.antisense)
        )

        intergenic = (
                r'''%s subtract -a <(cat %s %s %s | awk '{if(match($0, /transcript_id "[a-zA-Z0-9]*.*[a-zA-Z0-9]*"/)) print substr($0, RSTART, RLENGTH)}' | sort -uV | grep -Fwvf - %s) -b %s > %s''' %
                (bedtools, self.intronic, self.sense, self.antisense, self.lncRNA_bed, self.gene_bed, self.intergenic)
        )

        return '\n'.join([gff2gtf, gene_bed, exon_bed, intron_bed, lncRNA_bed, sense, antisense, intronic, intergenic])



"""
File: biopandas.py
Description: Common functions to handle dataframe.
CreateDate: 2022/1/10
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import Union, List, Iterable
from io import StringIO
from warnings import filterwarnings
from re import sub
from click import echo, open_file
from pandas import cut, set_option, read_table, read_excel, read_csv, Series, DataFrame, ExcelWriter


def display_set(decimal: int = 2) -> None:
    filterwarnings("ignore")
    set_option('display.float_format', lambda x: f'%.{decimal}f' % x)
    set_option('display.width', 1000)
    set_option('display.max_columns', None)
    # pd.set_option('display.max_rows', None)


def read_file_as_dataframe_from_stdin(sep: str = '\t',
                                      line_terminator: str = '\n',
                                      index_col: Union[int, None] = 0,
                                      header: int = 0,
                                      skip_rows: int = 0,
                                      use_cols: List[Union[str, int]] = None,
                                      names: List[str] = None,
                                      chunk_size: int = None):
    df = read_table(StringIO(''.join(open_file('-').readlines())),
                    sep=sep, header=header, names=names, index_col=index_col, usecols=use_cols,
                    lineterminator=line_terminator, skiprows=skip_rows, chunksize=chunk_size)
    return df


def read_in_gene_expression_as_dataframe(gene_exp_file: str) -> Union[str, DataFrame]:
    function_dict = {'txt': read_table, 'xlsx': read_excel, 'xls': read_excel, 'csv': read_csv}
    try:
        df = function_dict[gene_exp_file.split('.')[-1]](gene_exp_file, index_col=0)
    except KeyError:
        return '\033[31mError: Unrecognised file formats. Only txt, xls, xlsx, and csv formats are supported.\033[0m'
    else:
        return df


def merge_duplicate_indexes(df: DataFrame) -> DataFrame:
    """Sums the column values of rows with the same index"""
    index_name = df.index.name
    data_frame = df.groupby(index_name).sum()
    return data_frame


def filter_by_min_value(df: DataFrame,
                        min_value: float,
                        start_column_num: int = 1,
                        end_column_num: int = None) -> DataFrame:
    """Delete rows where all column values are less than the specified value"""
    if end_column_num is None:
        data_frame = df[~(df.iloc[:, start_column_num - 1:] <= min_value).all(axis=1)]
        return data_frame
    else:
        data_frame = df[~(df.iloc[:, start_column_num - 1:end_column_num - 1] <= min_value).all(axis=1)]
        return data_frame


def get_FPKM(read_count: DataFrame,
             min_value: float = None) -> DataFrame:
    """
    Standardize gene expression with FPKM.
    :param read_count: read count DataFrame. (type=pd.DataFrame)

            Geneid     Length   Sample1   Sample2   Sample3
            gene1      200         1         0         0
            gene2      1000       21        51        34
            ......     .......   .......   .......   .......
            gene1000   2100      2345      2137      1987

    :param min_value: Gene minimum expression (genes whose expression is less than the specified value in all samples
                      are filtered out).
    :return: FPKM matrix. (type=pd.DataFrame)
    """
    read_count_DataFrame = merge_duplicate_indexes(read_count)
    div_gene_length = read_count_DataFrame.div(read_count_DataFrame.iloc[:, 0], axis=0)
    div_gene_length.loc['sum'] = read_count_DataFrame.sum(axis=0)
    div_read_num = div_gene_length.div(div_gene_length.loc['sum'].T, axis=1) * 10 ** 9
    FPKM = div_read_num.iloc[0:-1, 1:]
    raw_gene_num = len(FPKM.index.tolist())
    if min_value:
        FPKM = filter_by_min_value(FPKM, min_value=min_value)
        new_gene_num = len(FPKM.index.tolist())
        echo(f"{raw_gene_num - new_gene_num} genes have been filtered out", err=True)
    return FPKM


def get_TPM(read_count: DataFrame,
            min_value: float = None) -> DataFrame:
    """
    Standardize gene expression with TPM.
    :param read_count: read count DataFrame. (type=pd.DataFrame)

            Geneid     Length   Sample1   Sample2   Sample3
            gene1      200         1         0         0
            gene2      1000       21        51        34
            ......     .......   .......   .......   .......
            gene1000   2100      2345      2137      1987

    :param min_value: Gene minimum expression (genes whose expression is less than the specified value in all samples
                      are filtered out).
    :return: TPM matrix. (type=pd.DataFrame)
    """
    read_count_DataFrame = merge_duplicate_indexes(read_count)
    read_count_DataFrame = read_count_DataFrame.div(read_count_DataFrame.iloc[:, 0], axis=0)
    read_count_DataFrame.loc['sum'] = read_count_DataFrame.sum(axis=0)
    read_count_DataFrame = read_count_DataFrame.div(read_count_DataFrame.iloc[-1], axis=1)
    read_count_DataFrame = read_count_DataFrame * 10 ** 6
    TPM = read_count_DataFrame.iloc[:-1, 1:]
    raw_gene_num = len(TPM.index.tolist())

    if min_value:
        TPM = filter_by_min_value(TPM, min_value=min_value)
        new_gene_num = len(TPM.index.tolist())
        echo(f"{raw_gene_num - new_gene_num} genes have been filtered out", err=True)
    return TPM


def dfs_to_excel(df_list: list, output_file: str, sheets_name: list = None):
    with ExcelWriter(output_file, engine='xlsxwriter') as writer:
        if not sheets_name:
            sheets_name = [f'Sheet {i}' for i in range(1, len(df_list) + 1)]
        workbook = writer.book
        workbook.nan_inf_to_errors = True  # 添加 'nan_inf_to_errors' 选项
        default_cell_format = workbook.add_format({'bold': False, 'border': 0, 'align': 'left'})
        for df, sheet_name in zip(df_list, sheets_name):
            df.fillna('NA', inplace=True)
            df = df.reset_index()
            df.to_excel(writer, sheet_name=sheet_name, na_rep='NA', startrow=1, header=False, index=False)
            worksheet = writer.sheets[sheet_name]  # 获取工作表对象
            # 将标题写入工作表，并应用默认的单元格格式
            for col_num, value in enumerate(df.columns.values):
                worksheet.write(0, col_num, value, default_cell_format)
            # 将数据写入工作表，并应用默认的单元格格式
            for row_num, row_data in enumerate(df.values):
                for col_num, col_data in enumerate(row_data):
                    worksheet.write(row_num + 1, col_num, col_data, default_cell_format)


def dataframe_to_str(df: DataFrame,
                     index: bool = True,
                     header: bool = True):
    if index:
        df = df.reset_index()
    string_df = sub(r'\n +', '\n', df.to_string(index=False, header=header).strip())
    string_df = sub(r' +', '\t', string_df)
    return string_df


def interval_stat(ser: Series, bins: Iterable, precision: int = 3, name: str = 'count'):
    ret: DataFrame = cut(x=ser, bins=bins, precision=precision, include_lowest=True).value_counts(sort=False).to_frame(name)
    ret.index.name = ret.index.name + '_interval'
    ret.rename(index={ret.index[0]: sub(r'.+,', '[0,', str(ret.index[0]))}, inplace=True)
    ret.loc['total', name] = ret[name].sum()
    return ret



#!/usr/bin/env python
"""
File: show_info.py
Description: Show other information.
CreateDate: 2023/8/4
Author: xuwenlin
E-mail: wenlinxu.njfu@outlook.com
"""
from typing import List
from click import echo


class Displayer:
    def __init__(self,
                 program: str,
                 authors: List[str] = ['Wenlin Xu'],
                 contacts: List[str] = ['wenlinxu.njfu@outlook.com'],
                 version: str = '1.1.0'):
        self.program = program
        self.authors = ', '.join(authors)
        self.contacts = ', '.join(contacts)
        self.version = version

    def version_info(self, ctx, param, value):
        if not value or ctx.resilient_parsing:
            return
        echo(f'Program: {self.program}\n'
             f'Author:  {self.authors}\n'
             f'Contact: {self.contacts}\n'
             f'Version: {self.version}',
             err=True)
        ctx.exit()
